#ifndef _ANALYZER_HPP_
#define _ANALYZER_HPP_

//メソッドアナライザー
//メソッドのcpu時間や使用回数など諸々を調べる

//part...どれかが単体で動くことを想定
//phase...一度の処理の中で連続して存在することを想定

//簡単に使うため、アナライザクラスには解析可能性のあるデータについて全て実装しておく
//最適化で邪魔な部分が消えてくれない限り、相当なメモリの無駄だが
//本番用では全体をオフにするので不問とする

//staticなものはスレッドセーフではないので並列化の際には注意

namespace Analysis{
	//アナライザーの表示タイプ
	
	enum{
		TYPE_PLANE=0,//回数、時間計測、失敗回数のみ
		TYPE_SEARCH,//探索結果
		TYPE_PLAYOUT,//未実装
	};
	
}

//アナライザと呼ぶほどもでない静的カウンタ、時間計測器

struct StaticCounter{
#ifdef USE_ANALYZER
	uint64 c;
#endif
	
	StaticCounter()
#ifdef USE_ANALYZER
	:c(0)
#endif
	{}
	
	StaticCounter(uint64 argC)
#ifdef USE_ANALYZER
	:c(argC)
#endif
	{}
	
	~StaticCounter(){
#ifdef USE_ANALYZER
		COUT<<"COUNTER : "<<c<<endl;	
#endif
	}
	
	void add(){
#ifdef USE_ANALYZER
		++c;
#endif
	}
};

struct StaticClock{
#ifdef USE_ANALYZER
	Clock cl;
	uint64 c;
	uint64 t;
#endif
	
	StaticClock()
#ifdef USE_ANALYZER
	:c(0U),t(0ULL)
#endif
	{}
	
	StaticClock(uint64 argC,uint64 argT)
#ifdef USE_ANALYZER
	:c(argC),t(argT)
#endif
	{}
	
	~StaticClock(){
#ifdef USE_ANALYZER
		COUT<<"CLOCK : "<<t<<" clock ( "<<c<<" times)";
		if( c>0 ){
			uint64 avg=t/c;
			COUT<<" "<<avg<<" per trial.";
		}
		COUT<<endl;
#endif
	}
	
	void start(){
#ifdef USE_ANALYZER
		cl.start();
#endif
	}
	void stop(){
#ifdef USE_ANALYZER
		uint64 tmp=cl.stop();
		++c;
		t+=tmp;
#endif
	}
};


template<int PART_NUM=1,int PHASE_MAX=1,int MODE=0>
struct StaticAnalyzer{
	//静的なやつ 手軽
#ifdef USE_ANALYZER
	Clock clock;
	const std::string& name;
	
	uint64 trials[PART_NUM];
	//時間
	uint64 time[PART_NUM][PHASE_MAX];
	//失敗
	uint64 failures[PART_NUM];
	//探索解析
	uint64 nodes[PART_NUM];
	uint64 childs[PART_NUM];
	//プレイアウト解析
	uint64 turns[PART_NUM];
#endif
	
	void startClock(){
#ifdef USE_ANALYZER
		clock.start();
#endif
	}
	
	uint64 stopClock(){
#ifdef USE_ANALYZER
		return clock.stop();
#else
		return 0ULL;
#endif
	}
	
	uint64 restartClock(){
#ifdef USE_ANALYZER
		return clock.restart();
#else
		return 0ULL;
#endif
	}

	void addTrial(const int pa=0){
#ifdef USE_ANALYZER
		++trials[pa];
#endif
	}
	
	void addTime(const uint64 argTime,const int pa=0,const int ph=0){
#ifdef USE_ANALYZER
		time[pa][ph]+=argTime;
#endif
	}

	void addFailure(const int pa=0){
#ifdef USE_ANALYZER
		++failures[pa];
#endif
	}
	
	void addNodes(const int argNodes,const int pa=0){
#ifdef USE_ANALYZER
		nodes[pa]+=argNodes;
#endif
	}
	
	void addChilds(const int argChilds,const int pa=0){
#ifdef USE_ANALYZER
		childs[pa]+=argChilds;
#endif	
	}
	
	void addTurns(const int argTurns,const int pa=0){
#ifdef USE_ANALYZER
		turns[pa]+=argTurns;	
#endif
	}
	
	void start(){
		startClock();
	}
	
	void restart(uint64 argTime,int pa,int ph){//外部クロックの場合
		addTime(argTime,pa,ph);
	}
	
	void restart(int pa=0,int ph=0){
		addTime(stopClock(),pa,ph);
		startClock();
	}
	
	void end(uint64 argTime,int pa,int ph){//外部クロックの場合
		addTime(argTime,pa,ph);
		addTrial(pa);
	}
	
	void end(int pa=0,int ph=0){
		end(stopClock(),pa,ph);
	}
	
	void init(){
#ifdef USE_ANALYZER
		for(int i=0;i<PART_NUM;++i){
			trials[i]=0U;
			failures[i]=0U;
			for(int j=0;j<PHASE_MAX;++j){
				time[i][j]=0ULL;
			}
		}
#endif
	}
	
	StaticAnalyzer(){init();}
	
	StaticAnalyzer(const std::string& argName)
#ifdef USE_ANALYZER	
	:name(argName)
#endif
	{
		init();
	}
	
	~StaticAnalyzer(){
		//結果表示
		
#ifdef USE_ANALYZER
		uint64 time_all_sum=0ULL;
		uint64 time_part_sum[PART_NUM]={0ULL};
		uint64 time_per_trial_all=0ULL;
		uint64 time_per_trial_part[PART_NUM]={0ULL};
		uint64 time_per_trial_phase[PART_NUM][PHASE_MAX]={0ULL};
		
		uint64 trial_sum=0U;
		uint64 failure_sum=0U;
		
		//sum
		for(auto i=0;i<PART_NUM;++i){
			if(trials[i]){
				trial_sum+=trials[i];
				failure_sum+=failures[i];
				for(int j=0;j<PHASE_MAX;++j){
					time_part_sum[i]+=time[i][j];
					time_per_trial_phase[i][j]=time[i][j]/trials[i];
				}
				time_all_sum+=time_part_sum[i];
				time_per_trial_part[i]=time_part_sum[i]/trials[i];
			}
		}
		if(trial_sum){
			time_per_trial_all=time_all_sum/trial_sum;
		}
		
		COUT<<"****** Analysis of "<<name<<" ******"<<endl;
		
		COUT<<" < Normal Analysis >"<<endl;
		
		COUT<<" "<<trial_sum<<" trials.  "<<time_all_sum<<" clock ( "<<time_per_trial_all<<" clock/trial).  "<<failure_sum<<" failures."<<endl;
		if(PART_NUM>1 || PHASE_MAX>1){
			for(auto i=0;i<PART_NUM;++i){
				COUT<<"   Part "<<i<<" : "<<trials[i]<<" trials.  "<<time_part_sum[i]<<" clock ( "<<time_per_trial_part[i]<<" clock/trial).  "<<failures[i]<<" failures."<<endl;
				if(PHASE_MAX > 1){
					for(auto j=0;j<PHASE_MAX;++j){
						COUT<<"      Phase "<<j<<" : "<<time[i][j]<<" clock ( "<<time_per_trial_phase[i][j]<<" clock/trial)."<<endl;
					}
				}
			}
		}
		
		if(MODE == Analysis::TYPE_SEARCH){
			uint64 node_sum=0U;
			uint64 child_sum=0U;
			
			uint64 node_per_trial_part[PART_NUM]={0U};
			uint64 node_per_trial_all=0U;
			
			uint64 time_per_node_part[PART_NUM]={0ULL};
			uint64 time_per_node_all=0ULL;
			
			uint64 child_per_trial_part[PART_NUM]={0U};
			uint64 child_per_trial_all=0U;
			
			uint64 time_per_child_part[PART_NUM]={0ULL};
			uint64 time_per_child_all=0ULL;
			
			for(auto i=0;i<PART_NUM;++i){
				if(trials[i]){
					node_sum+=nodes[i];
					child_sum+=childs[i];
					
					if(nodes[i]){
						node_per_trial_part[i]=nodes[i]/trials[i];
						time_per_node_part[i]=time_part_sum[i]/nodes[i];
					}
						
					if(childs[i]){
						child_per_trial_part[i]=childs[i]/trials[i];
						time_per_child_part[i]=time_part_sum[i]/childs[i];
					}
				}
			}
			
			if(trial_sum){
				if(node_sum){
					node_per_trial_all=node_sum/trial_sum;
					time_per_node_all=time_all_sum/node_sum;
				}
				if(child_sum){
					child_per_trial_all=child_sum/trial_sum;
					time_per_child_all=time_all_sum/child_sum;
				}
			}
			
			
			COUT<<" < Search Analysis >"<<endl;
			COUT<<" "<<node_sum<<" nodes ( "<<node_per_trial_all<<" nodes/trial ; "<<time_per_node_all<<" clock/node).  ";
			COUT<<" "<<child_sum<<" childs ( "<<child_per_trial_all<<" childs/trial ; "<<time_per_child_all<<" clock/child).  "<<endl;
			if(PART_NUM>1){
				for(auto i=0;i<PART_NUM;++i){
					COUT<<"   Part "<<i<<" : "<<nodes[i]<<" nodes ( "<<node_per_trial_part[i]<<" nodes/trial ; "<<time_per_node_part[i]<<" clock/node).  ";
					COUT<<childs[i]<<" childs ( "<<child_per_trial_part[i]<<" childs/trial ; "<<time_per_child_part[i]<<" clock/child)."<<endl;
				}
			}
		}
#endif
	}
};

struct HashBookAnalyzer{
	//置換表の解析用
	
#ifdef USE_ANALYZER	
	const std::string name;
	const uint64 book_size;
	
	uint64 fill;
	uint64 hit;
	uint64 diffHash;
	uint64 white;
	uint64 regist;
#endif
	
	void init(){
#ifdef USE_ANALYZER	
		fill=0;hit=0;diffHash=0;white=0;regist=0;
#endif
	}
	
	void addFill(){
#ifdef USE_ANALYZER
		++fill;
#endif
	}
	
	void addHit(){
#ifdef USE_ANALYZER
		++hit;
#endif
	}
	
	void addDiffHash(){
#ifdef USE_ANALYZER
		++diffHash;
#endif	
	}
	
	void addWhite(){
#ifdef USE_ANALYZER
		++white;
#endif		
	}
	
	void addRegist(){
#ifdef USE_ANALYZER
		++regist;	
#endif
	}
	
	//HashBookAnalyzer(){init();}
	
	HashBookAnalyzer(const std::string& argName,const int size)
#ifdef USE_ANALYZER	
	:name(argName),book_size(size)
#endif
	{
		init();
	}
	
	~HashBookAnalyzer(){
#ifdef USE_ANALYZER	
		double realRate=fill/(double)book_size;
		double idealRate=1.0-pow((1.0-1.0/(double)book_size),regist);
		COUT<<"****** Analysis(HashBook) of "<<name<<" ******"<<endl;
		COUT<<"   fill Rate : "<<realRate<<" ( ideally... "<<idealRate<<" )  "<<"Registration: "<<regist<<" / "<<book_size<<endl;
		COUT<<"   hit : "<<hit<<"  diffHash : "<<diffHash<<"  white : "<<white<<endl;
#endif
	}
};


#endif /* _ANALYZER_HPP_ */