//デジタルカーリング
//モンテカルロ着手決定機関

namespace DigitalCurling{
	namespace Ayumu{
		class MonteCarloPlayer{
		private:
		public:
			
			template<class board_t>
			uint64 decide_play_time(const board_t& bd);
			
			template<class board_t>
			int play(const board_t& bd,MoveXY *const dst);
			
			template<class board_t>
			fpn_t eval(const board_t& bd);
			
			MonteCarloPlayer(){}
			
			~MonteCarloPlayer(){}
		};
		
		template<class board_t>
		uint64 MonteCarloPlayer::decide_play_time(const board_t& bd){
			//自分の着手決定限度時間を設定
			uint64 tmp_time_limit;
			const auto& my_tl=bd.tl[bd.my_first_color];
			
			if( my_tl.isUnlimited() ){//持ち時間無制限
				//計算量レベルに合わせて決定する予定
				tmp_time_limit=10000;//とりあえず10秒
			}else{
				const uint64 insurance=2000;
				const uint64 insurance_by_shot=400;
				const uint32 shots_remained=bd.getNRemainedEnds() * 8 + (bd.getNRemainedTurns()/2 +1);
				//エンド、ターンから決定
				if( my_tl.limit_ms <= insurance_by_shot * shots_remained + insurance ){
					tmp_time_limit = my_tl.limit_ms / shots_remained;
				}else{
					tmp_time_limit = ( my_tl.limit_ms - insurance_by_shot * shots_remained - insurance )/shots_remained;
				}
			}
			
			return tmp_time_limit;
		}
		
		template<class board_t>
		int MonteCarloPlayer::play(const board_t& bd,MoveXY *const dst){
			//モンテカルロで着手決定
			typedef MoveXY move_t;
			
			RootNode<move_t> mcf;
			ThinkBoard tb;
			
			tb.set(bd);//局面情報を思考用にセット
			
			//候補手(言語型)の生成
			KillerMove kmv[64];
			int NKillers=0;
			
			//定跡から手の候補を入れる
			NKillers=pushKillerRoot( kmv, tb );
			
			//まず定跡からの決定手があればそれに決定
			for(int i=0;i<NKillers;++i){
				if( kmv[i].isSettled() ){
					fMoveXY<> fmv;
					genKiller(tb,kmv[i],&fmv);
					convMove(fmv,dst);
					return 1;
				}
			}
			
			//MCFieldに候補着手を用意
			mcf.init();
			move_t mv_s0,mv_s1;
			mv_s0.setWR(0);
			mv_s1.setWR(1);
			
			move_t tmpm;
			fRMoveXY<> m;
			
			//定跡のレンジ手を入れる
			for(int i=0;i<NKillers;++i){
				genRKiller(tb,kmv[i],&m);
				convRMove(m,&tmpm);
				mcf.addChild(tb,tmpm);
			}
			
			if( tb.isSymmetry() ){
				//左右対称の時は、Vxが正または負の着手のみ考えれば良い
				DERR<<"Sym"<<endl;
				if( thread_tools[0].ddice.rand()%2==0 ){
					for(int i=0;i<4;++i){
						tmpm=mv_s0.part(4,i);
						mcf.addChild(tb,tmpm);
					}
				}else{
					for(int i=0;i<4;++i){
						tmpm=mv_s1.part(4,i);
						mcf.addChild(tb,tmpm);
					}	
				}
			}else{
				for(int i=0;i<4;++i){
					tmpm=mv_s0.part(4,i);
					mcf.addChild(tb,tmpm);
					tmpm=mv_s1.part(4,i);
					mcf.addChild(tb,tmpm);
				}
			}
			
			//もしこの時点で候補着手が無い場合、ティーショットをする
			if( mcf.childs<=0 ){
				genTeeShot(&m);
				convMove(m,dst);
				return 0;
			}
			
			//時間制限を決める
			uint64 tmp_time_limit=decide_play_time(bd);
			
			CERR<<"thinking time limit ... "<<tmp_time_limit<<" ms"<<endl;
			
			mcf.set_limit_time( tmp_time_limit );
			
			//報酬系設定
			setTmpEndReward(tb,tb.getTurnColor());
			
#ifdef MULTI_THREADING
			
			//マルチスレッディングの場合、各スレッドを起動
			thread th0(mc_thread<RootNode<move_t>>,0,&mcf,&tb);
#if N_THREADS >= 2
			thread th1(mc_thread<RootNode<move_t>>,1,&mcf,&tb);
#endif
#if N_THREADS >= 3
			thread th2(mc_thread<RootNode<move_t>>,2,&mcf,&tb);
#endif
#if N_THREADS >= 4
			thread th3(mc_thread<RootNode<move_t>>,3,&mcf,&tb);
#endif
#if N_THREADS >= 5
			thread th4(mc_thread<RootNode<move_t>>,4,&mcf,&tb);
#endif
#if N_THREADS >= 6
			thread th5(mc_thread<RootNode<move_t>>,5,&mcf,&tb);
#endif
#if N_THREADS >= 7
			thread th6(mc_thread<RootNode<move_t>>,6,&mcf,&tb);
#endif
#if N_THREADS >= 8
			thread th7(mc_thread<RootNode<move_t>>,7,&mcf,&tb);
#endif
#if N_THREADS >= 9
			static_assert(0,"num of threads over.");
#endif
			
			th0.join();
#if N_THREADS >= 2
			th1.join();
#endif
#if N_THREADS >= 3
			th2.join();
#endif
#if N_THREADS >= 4
			th3.join();
#endif
#if N_THREADS >= 5
			th4.join();
#endif
#if N_THREADS >= 6
			th5.join();
#endif
#if N_THREADS >= 7
			th6.join();
#endif
#if N_THREADS >= 8
			th7.join();
#endif
			
#else
			//関数として開く
			mc_thread<RootNode<move_t>>(0,&mcf,&tb);
			
#endif //MULTI_THREADING
			
			int best_id=mcf.choose_best_id();//最高評価のもの
			MoveXY bestMove=MoveXY(mcf.child[best_id].dive_largest());
			
#ifdef MONITOR
			CERR<<"MCP : "<<mcf.trials<<" playouts were done."<<endl;
			CERR<<"MCP : "<<mvTT.npages<<" children were made."<<endl;
			CERR<<"MCP : "<<bdTT.npages<<" nodes were made."<<endl;
			
			//bdTT.print();
			
			mcf.print();
			
			CERR<<"MCP : best move - "<<bestMove<<endl;
			
#endif //MONITOR
			
			MoveXY bestNRMove=bestMove.genNR(&thread_tools[0].ddice);
			
#ifdef MONITOR
			CERR<<"MCP : best no-range move - "<<bestNRMove<<endl;
#endif //MONITOR
			
			//getchar();
			
			*dst=bestMove.genNR(&thread_tools[0].ddice);
			return 0;
		}
		
		template<class board_t>
		fpn_t MonteCarloPlayer::eval(const board_t& bd){
			//モンテカルロで局面評価
			//限度時間を決め、その時間一杯まで考える
			typedef MoveXY move_t;
			
			
			return 0;
		}
	}
}