//デジタルカーリング用
//高速シミュレーション
//局面引数の基本型はThinkBoard

//用語

//V ストーンの速さ
//R ストーンが衝突無しに静止する地点までの距離
//T ストーンが衝突無しに静止するまでの時間

//他の注意点

//座標はティーを(0,0)として、センターラインの投げる向きをy軸、その正方向に対して右向きをx軸としている

//極座標での角度θは通常はx=r*cosθ、y=r*sinθと定義されるが、
//本プログラムにおいてはx=r*sinθ、y=r*cosθとしている。
//こうすることでセンターライン対称が角度の符号を変える事で表現出来る

namespace DigitalCurling{
	namespace FastSimulator{
		
		//定数
		constexpr fpn_t FTHETA_CURL_MAX=0.03333565;//停止位置への回転最大角度
		constexpr fpn_t FR_COLLISION=0.0006;//衝突閾値距離
		constexpr fpn_t F_SLID_STONES=0.046;//石どうしの滑り係数
		
		//T <-> V 変換
		//現在の物理モデルではTとVは比例なので、直接計算
		
		//T_V変換
		constexpr fpn_t FTtoFV(fpn_t ft){ return F_FRIC_STONE * ft; }
		//V_T変換
		constexpr fpn_t FVtoFT(fpn_t fv){ return fv / F_FRIC_STONE; }
		
		//V <-> R 変換
		//現在の物理モデルではこの変換の直接計算は難しそうなので、テーブルを利用して変換
		//テーブルサイズ縮小のため、端数を線形に埋めるひと手間をかけている
		
		//V_R変換
		constexpr int FV_FR_TABLE_RESOLUTION=14;
		std::array<fpn_t,(1<<FV_FR_TABLE_RESOLUTION)+4> FV_FRTable;
		constexpr fpn_t FV_FR_TABLE_STEP=(FV_MAX-0)/fpn_t(1<<FV_FR_TABLE_RESOLUTION);
		
		fpn_t FVtoFR(fpn_t fv){
			uint32 fid=uint32(fv / FV_FR_TABLE_STEP);
			ASSERT( 0<=fid && fid<(1<<FV_FR_TABLE_RESOLUTION)+4, cerr<<fv/FV_MAX<<endl; );
			fpn_t hasu=fv - ((fpn_t)fid) * FV_FR_TABLE_STEP;
			ASSERT( 0<=hasu && hasu<=1.0, cerr<<hasu<<endl; );
			return FV_FRTable[fid]*(1-hasu) + FV_FRTable[fid+1]*hasu;
		}
		
		//R_V変換
		constexpr int FR_FV_TABLE_RESOLUTION=14;
		std::array<fpn_t,(1<<FR_FV_TABLE_RESOLUTION)+4> FR_FVTable;
		constexpr fpn_t FR_FV_TABLE_STEP=(49.0)/fpn_t(1<<FR_FV_TABLE_RESOLUTION);
		
		fpn_t FRtoFV(fpn_t fr){
			int fid=int(fr / FR_FV_TABLE_STEP);
			ASSERT( 0<=fid && fid<(1<<FR_FV_TABLE_RESOLUTION)+4, cerr<<fr/49.0<<endl; );
			fpn_t hasu=fr - ((fpn_t)fid) * FR_FV_TABLE_STEP;
			ASSERT( 0<=hasu && hasu<=1.0, cerr<<hasu<<endl; );
			return FR_FVTable[fid]*(1-hasu) + FR_FVTable[fid+1]*hasu;
		}
		
		//T <-> R 変換
		//T<->VとV<->Rの変換の合成
		
		//T_R変換
		fpn_t FTtoFR(fpn_t ft){ return FVtoFR(FTtoFV(ft)); }
		//R_T変換
		fpn_t FRtoFT(fpn_t fr){ return FVtoFT(FRtoFV(fr)); }
		
		
		//シミュレーションのメイン変換
		//V(Tでも同じ事だが)からの変換によって石の状態を読む
		
		//格納データ
		//停止位置を(0,0)とし、停止直前にはy軸正の向きに動いていたストーンについて、
		//それぞれの時点での位置座標(x,y)と石の向き、
		//時刻(速さから計算出来るので持っておく意味は少ないと思う)		
		struct FSimuData{
			//必要なデータ
			fpn_t x,y;
			fpn_t theta;
			fpn_t t;
			
			void set(fpn_t ax,fpn_t ay,fpn_t atheta,fpn_t at){
				x=ax;y=ay;theta=atheta;t=at;
			}
			
			fpn_t getX()const{return x;}
			fpn_t getY()const{return y;}
			fpn_t getTheta()const{return theta;}
			fpn_t getTime()const{return t;}
		};
		
		constexpr int FV_TABLE_RESOLUTION=18;
		std::array<FSimuData,(1<<FV_TABLE_RESOLUTION)+4> V_ALLTable;
		constexpr fpn_t FV_TABLE_STEP=(FV_MAX+2.0)/fpn_t(1<<FV_TABLE_RESOLUTION);
		
		constexpr uint32 FVtoID(fpn_t fv){
			return (uint32)(fv/FV_TABLE_STEP);
		}
		constexpr fpn_t IDtoFV(uint32 id){
			return ((fpn_t)id)*FV_TABLE_STEP;
		}
		const FSimuData& FVtoFALL(fpn_t fv){
			return V_ALLTable[FVtoID(fv)];
		}
		
		//T_ALL変換
		constexpr uint32 FTtoID(fpn_t ft){ return FVtoID(FTtoFV(ft)); }
		constexpr fpn_t IDtoFT(uint32 id){ return FVtoFT(IDtoFV(id)); }
		const FSimuData& FTtoFALL(fpn_t ft){ return V_ALLTable[FTtoID(ft)]; }
		
		//R_ALL変換
		uint32 FRtoID(fpn_t fr){ return FVtoID(FRtoFV(fr)); }
		fpn_t IDtoFR(uint32 id){ return FVtoFR(IDtoFV(id)); }
		
		
		//シミュレーション中の石の状態を表すための構造体
		struct FMoment{
			fpn_t x,y;
			fpn_t theta;
			fpn_t t;
			fpn_t v,w;
			
			void set(fpn_t ax,fpn_t ay,fpn_t atheta,fpn_t at,fpn_t av,fpn_t aw){
				x=ax;y=ay;theta=atheta;t=at;v=av;w=aw;
			}
			fpn_t getX()const{return x;}
			fpn_t getY()const{return y;}
			fpn_t getVX()const{return v * sin(theta);}
			fpn_t getVY()const{return v * cos(theta);}
			fpn_t getW()const{return w;}
		};
		
		struct FSMovingStone : public FMoment{
			BitSet32 asleep;//自分と衝突可能性がある静止石のデータ
			BitSet32 awake;//自分と衝突可能性がある動石のデータ
			fpn_t r;//_goal;//停止位置までの距離
			fpn_t t_goal;
			StoneState state;
		};
		
		struct FSBoard{
			FSMovingStone stone[16];//石の状態(順番に意味は無い)
			//動いている石
			BitSet32 awake;
			int NAwake;
			//止まっている石
			BitSet32 asleep;
			int NAsleep;
			int col;//手番を持っている色
			
			int getTurnColor()const{return col;}
			
			template<class board_t,class mst_t>
			void setAfterFirstColl(const board_t& bd,const int as_coll, const mst_t& aw0, const mst_t& aw1){
				//最初の衝突後にデータをセット
				//動いている石を0番と1番の石とする
				stone[0].set( aw0.x, aw0.y, aw0.theta, aw0.t, aw0.v, aw0.w );
				stone[1].set( aw1.x, aw1.y, aw1.theta, aw1.t, aw1.v, aw1.w );
				
				awake.reset();
				awake.set(0);awake.set(1);
				NAwake=2;
				
				col=bd.getTurnColor();
				
				asleep=0U;
				int cnt=NAwake;
				
				//静止している石を設定
				for(int c=0;c<2;++c){
					for(int i=0;i<bd.NActive[c];++i){
						if( as_coll != c*8+i ){
							asleep.set(cnt);
							stone[cnt].x=bd.stone[c][i].x;
							stone[cnt].y=bd.stone[c][i].y;
							stone[cnt].state=bd.state[c][i];
							
							DERR<<bd.stone[c][i]<<endl;
							++cnt;
						}
					}
				}
				
				NAsleep=cnt-NAwake;
				
				//動いている石のそれぞれに対して、衝突可能な静止石の集合をセット
				//現在では全ての静止石をセットするだけ
				stone[0].asleep=asleep;
				stone[1].asleep=asleep;
				
				stone[0].awake.reset();
				stone[1].awake.reset();
				
				//石の付加情報
				stone[0].state.init();
				stone[0].state.setColor(col);
				
				stone[1].state=bd.state[as_coll/8][as_coll%8];
			}
			bool exam()const{//validator
				
				//awake,asleep list
				if( !isExclusive((uint32)awake,(uint32)asleep) ){
					cerr<<"FSBoard::exam() : awakes and asleeps inexclusive."<<endl;
					return false;
				}
				for( BitSet32 tmp=awake; tmp.any(); tmp.pop_lsb()){
					int aw=tmp.bsf();
					if( !awake.holds(stone[aw].awake) ){
						cerr<<"FSBoard::exam() : all awakes doesn't hold aw_stone("<<aw<<")'s awakes."<<endl;
						return false;
					}
					if( !asleep.holds(stone[aw].asleep) ){
						cerr<<"FSBoard::exam() : all asleeps doesn't hold aw_stone("<<aw<<")'s asleeps."<<endl;
						return false;
					}
					if( stone[aw].awake.test(aw) ){
						cerr<<"FSBoard::exam() : aw_stone("<<aw<<")'s awakes holds myself."<<endl;
						return false;
					}
					if( stone[aw].asleep.test(aw) ){
						cerr<<"FSBoard::exam() : aw_stone("<<aw<<")'s asleeps holds myself."<<endl;
						return false;
					}
					if( !awake.test(aw) ){
						cerr<<"FSBoard::exam() : awakes doesn't hold aw_stone("<<aw<<")"<<endl;
						return false;
					}
					if( asleep.test(aw) ){
						cerr<<"FSBoard::exam() : asleeps holds aw_stone("<<aw<<")"<<endl;
						return false;
					}
				}
				//the number of stones
				int nst[2]={0};
				for( BitSet32 tmp=awake; tmp.any(); tmp.pop_lsb()){
					int aw=tmp.bsf();
					nst[stone[aw].state.getColor()]++;
				}
				for( BitSet32 tmp=asleep; tmp.any(); tmp.pop_lsb()){
					int as=tmp.bsf();
					nst[stone[as].state.getColor()]++;
				}
				for(int c=0;c<2;++c){
					if( nst[c]>8 ){
						cerr<<"FSBoard : illegal number of stones( "<<OutColor(c)<<nst[c]<<" )."<<endl;
						return false;
					}
				}
				return true;
			}
		};
		
		//初期化
		//通常の浮動小数点演算はfpn_tで行うが、
		//初期化時は高精度演算用の型であるpfpn_tで行う
		
		//初期化スレッド
		void initV_R(){//V_R変換初期化
			fMStone<pfpn_t> mst( 0, 0, 0, 0, 0 );
			for(int i=0;i<(1<<FV_FR_TABLE_RESOLUTION)+4;++i){
				//V_Rテーブルに現時点での値を入れていく
				FV_FRTable[i]=XYtoR( mst.x, mst.y );
				for(int j=0;j<256;++j){
					Simulator::invStepSolo<pfpn_t>( &mst, FV_FR_TABLE_STEP/F_FRIC_STONE/256 );
				}
			}
		}
		void initR_V(){//R_V変換初期化
			constexpr pfpn_t step=static_cast<pfpn_t>(1/700000.0);
			
			fMStone<> mst( 0, 0, 0, 0, 0 ),lmst,tmst;
			FR_FVTable[0]=0;
			pfpn_t tm=0;//逆算時刻
			for(int i=0;i<(1<<FR_FV_TABLE_RESOLUTION)+4;++i){
				//R_Vテーブルに現時点での値を入れていく
				pfpn_t line=i*FR_FV_TABLE_STEP;
				pfpn_t fr;
				while(1){
					fr=XYtoR(mst.x,mst.y);
					if( fr >= line ){break;}
					Simulator::invStepSolo<pfpn_t>( &mst, step );
					tm+=step;
				}
				tmst=mst;
				lmst=mst;
				Simulator::stepSolo( &lmst, step );
				pfpn_t lfr=XYtoR(lmst.x,lmst.y);
				pfpn_t hasu;
				if( fr!=lfr ){
					hasu=(line-lfr)/(fr-lfr);
				}else{
					hasu=static_cast<pfpn_t>(0.5);
				}
				ASSERT( 0<=hasu && hasu<=1.0, cerr<<double(lfr)<<"->"<<double(fr)<<"("<<double(line)<<")"<<endl;  );
				FR_FVTable[i]=XYtoR(lmst.vx,lmst.vy)*(1-hasu) + XYtoR(tmst.vx,tmst.vy)*hasu;
			}
		}
		void initALL(){//V_ALL変換初期化
			constexpr pfpn_t step=static_cast<pfpn_t>(1/700000.0);
			
			fMStone<pfpn_t> mst( 0, 0, 0, 0, 0 ),lmst,tmst;
			V_ALLTable[0].set( 0,0,0,0 );
			fpn_t tm=0;//逆算時刻
			for(int i=1;i<(1<<FV_TABLE_RESOLUTION)+4;++i){
				pfpn_t line=IDtoFV(i);
				pfpn_t fv;
				while(1){
					fv=XYtoR(mst.vx,mst.vy);
					if( fv >= line ){break;}
					Simulator::invStepSolo<pfpn_t>( &mst, step );
					tm+=step;
				}
				if( i%64 == 0 ){
					//cerr<<XYtoT(-mst.y,-mst.x)-XYtoT(mst.vy,mst.vx)<<endl;
				}
				tmst=mst;
				lmst=mst;
				Simulator::stepSolo<pfpn_t>( &lmst, step );
				pfpn_t lfv=XYtoR(lmst.vx,lmst.vy);
				pfpn_t hasu;
				if( fv!=lfv ){
					hasu=(line-lfv)/(fv-lfv);
				}else{
					hasu=static_cast<pfpn_t>(0.5);
				}
				ASSERT( 0<=hasu && hasu<=1.0, cerr<<double(lfv)<<"->"<<double(fv)<<"("<<double(line)<<")"<<endl;  );

				V_ALLTable[i].set(
								  -lmst.x*(1-hasu) + -tmst.x*hasu,
								  -lmst.y*(1-hasu) + -tmst.y*hasu,
								  -XYtoT(lmst.vy,lmst.vx)*(1-hasu) + -XYtoT(tmst.vy,tmst.vx)*hasu,
								  (tm-step)*(1-hasu) + tm*hasu
								  );
				}
		}
		
		void init(){
			
			//高精度演算による初期化は時間がかかるので、並列化している
			thread th0(initV_R);
			thread th1(initR_V);
			thread th2(initALL);
			
			th0.join();
			th1.join();
			th2.join();
		}
		
		template<class mst_t,class st_t>
		void collisionAw_As(mst_t *const aw,st_t *const as){
			//衝突処理(動-静)
			
			//衝突面がx軸と平行になるように回転
			fpn_t coltheta( XYtoT( as->y - aw->y, as->x - aw->x ) );
			
			fpn_t vn( aw->v * cos(aw->theta - coltheta) ),vt( aw->v * sin(aw->theta - coltheta) );
			
			//力積を質量で割ったもの
			fpn_t n_m( vn );
			fpn_t f_m( ( 1.0 - F_SLID_STONES ) / 6.0 * ( F_STONE_RAD * aw->w + vt ) );
			
			//角速度変化量
			fpn_t dw( f_m * 2.0 / F_STONE_RAD /* *1.056 */ );
			
			//静止していた石の状態更新
			as->v=XYtoR( n_m, f_m );
			as->theta=XYtoT( n_m, f_m ) + coltheta;
			as->w=dw;
			
			as->t=aw->t;
			
			//動いていた石の状態更新
			aw->v=XYtoR( vn-n_m, vt-f_m );
			aw->theta=XYtoT( vn-n_m, vt-f_m ) + coltheta;
			aw->w+=dw;
		}
		template<class mst_t>
		void collision2Aw(mst_t *const aw0,mst_t *const aw1){
			//衝突処理(動-動)
			
			//衝突面がx軸と平行になるように回転
			fpn_t coltheta( XYtoT( aw1->y - aw0->y, aw1->x - aw0->x ) );
			
			fpn_t v0n( aw0->v * cos(aw0->theta - coltheta) );
			fpn_t v0t( aw0->v * sin(aw0->theta - coltheta) );
			
			fpn_t v1n( aw1->v * cos(aw1->theta - coltheta) );
			fpn_t v1t( aw1->v * sin(aw1->theta - coltheta) );
			
			//力積を質量で割ったもの
			fpn_t n_m( v0n - v1n );
			fpn_t f_m( ( 1.0 - F_SLID_STONES ) / 6.0 * ( F_STONE_RAD * (aw0->w + aw1->w)  + (v0t-v1t) ) );
			
			//角速度変化量
			fpn_t dw( f_m * 2.0 / F_STONE_RAD /* * 1.04 */ );
			
			aw1->v=min( FV_MAX, XYtoR( v1n+n_m, v1t+f_m ) );
			
			ASSERT( aw1->v <= FV_MAX, cerr<<aw1->v<<endl; );
			
			aw1->theta=XYtoT( v1n+n_m, v1t+f_m ) + coltheta;
			aw1->w+=dw;
			
			aw0->v=min( FV_MAX, XYtoR( v0n-n_m, v0t-f_m ) );
			
			ASSERT( aw0->v <= FV_MAX, cerr<<aw0->v<<endl; );
			
			aw0->theta=XYtoT( v0n-n_m, v0t-f_m ) + coltheta;
			aw0->w+=dw;
		}
		
		template<class pos_t,class mv_t>
		void FXYtoFMV(const pos_t& pos,mv_t *const mv){
			//posで静止するVX,VYを計算
			fpn_t dx( pos.getX() - FX_THROW );
			fpn_t dy( pos.getY() - FY_THROW );	 
			fpn_t r( XYtoR(dx,dy) );
			fpn_t dtheta( XYtoT( dy, dx) );
			fpn_t v( FRtoFV( r ) );
			
			fpn_t dtx,vtheta;//テーブルでの変化量
			int id=FVtoID(v);
			if( !mv->isLeftSpin() ){
				dtx=+V_ALLTable[id].x;
				vtheta=+V_ALLTable[id].theta;
			}else{
				dtx=-V_ALLTable[id].x;
				vtheta=-V_ALLTable[id].theta;
			}
			
			fpn_t dttheta( XYtoT(V_ALLTable[id].y, dtx) );
			
			rotate( v, 0,
				   -dttheta + dtheta - vtheta,
				   &mv->y, &mv->x
				   );
		}
		
		template<class mst_t>
		void step(const uint32 id,const uint32 nid,mst_t *const dst){
			//前計算データから石の位置をステップさせる
			//id...現在状態のインデックス
			//nid...次状態のインデックス
			fpn_t 
			dx( V_ALLTable[id].x ),
			dy( V_ALLTable[id].y ),
			dt( V_ALLTable[id].t - V_ALLTable[nid].t ),
			dvtheta( V_ALLTable[id].theta - V_ALLTable[nid].theta ),
			dtheta( V_ALLTable[id].theta );
			
			dx-=V_ALLTable[nid].x;
			dy-=V_ALLTable[nid].y;
			
			if( !(dst->w > 0) ){
				dx=-dx;
				dvtheta=-dvtheta;
				dtheta=-dtheta;
			}
			rotateToAdd(
						dy,dx,
						dst->theta + dtheta,
						&dst->y,&dst->x
						);//位置を進める
			dst->theta+=dvtheta;//向き設定
			dst->t+=dt;//時刻設定
			dst->v-=dt * F_FRIC_STONE;//減速
		}
		
		template<class mst_t>
		void stepToStop(const uint32 id,mst_t *const dst){
			//前計算データから石の位置をステップさせる
			//衝突せずに静止するまで動かす
			//id...現在状態のインデックス
			fpn_t dx,dtheta;
			if( dst->w > 0 ){
				dx=V_ALLTable[id].x;
				dtheta=V_ALLTable[id].theta;
			}else{
				dx=-V_ALLTable[id].x;
				dtheta=-V_ALLTable[id].theta;
			}
			rotateToAdd(
						V_ALLTable[id].y, dx,
						dst->theta + dtheta,
						&dst->y, &dst->x
						);//座標設定
			dst->t+=V_ALLTable[id].t;//時刻設定
		}
		
		
		template<int RINK_ONLY=1,class board_t,class board2_t>
		int simulateF_b2d_FSB( board_t *const pbd, board2_t *const tobd, const fpn_t step )
		{
			//シミュレータ、box2dへの入り口
			//Awakeな石が複数あっていい場合
			
			b2Board b2bd;
			
			//ゲーム情報の初期化
			BitSet32 tmp;
			//awakeな石をセット
			tmp=pbd->awake;
			for( ;tmp.any();tmp.pop_lsb() ){
				int id=tmp.bsf();
				b2bd.setAwakeStone( pbd->stone[id], pbd->stone[id].state.getColor() );
			}
			//asleepな石をセット
			tmp=pbd->asleep;
			for( ;tmp.any();tmp.pop_lsb() ){
				int id=tmp.bsf();
				b2bd.setAsleepStone<RINK_ONLY>( pbd->stone[id], pbd->stone[id].state );
			}
			
			//物理演算
			//公式のものをベースとした関数で結果を求める
			int ret=Simulator::MainLoop<RINK_ONLY>( &b2bd ,step, -1 );
			
			assert( ret!=-1 );
			
			// シミュレーション結果をまとめる
			tobd->initStonePos();
			const int size=b2bd.NAwake+b2bd.NAsleep;
			
			for( int i=0; i<size; ++i ){
				assert( b2bd.body[i] != nullptr );
				const int col=b2bd.state[i].getColor();
				const b2Vec2 vec = b2bd.body[i]->GetPosition();
				if( (!RINK_ONLY) || isInPlayArea(vec.x,vec.y) ){
					tobd->pushStone(col,vec.x,vec.y);
				}
			}
			return ret;
		}
		
		template<int RINK_ONLY=1,class board_t,class move_t>
		int simulateSoloF(board_t *const pbd, const move_t& mv){
			//ただ停止位置を計算し、石を置くだけ
			//元々石がある場合にはおかしなことになりうるので、注意
			const int col=pbd->getTurnColor();
			//fPosXY<> dst;
			//calcDestination( mv, &dst );
			fpn_t v( XYtoR( mv.getVX(), mv.getVY() ) );
			fpn_t vtheta( XYtoT( mv.getVY(), mv.getVX() ) );
			fpn_t x(FX_THROW),y(FY_THROW),dx,dtheta;
			
			int id=FVtoID(v);
			if( !mv.isLeftSpin() ){
				dx=+V_ALLTable[id].x;
				dtheta=+V_ALLTable[id].theta;
			}else{
				dx=-V_ALLTable[id].x;
				dtheta=-V_ALLTable[id].theta;
			}
			rotateToAdd( V_ALLTable[id].y, dx, 
						vtheta + dtheta,
						&y, &x
						);
			
			if( (!RINK_ONLY) || isInPlayArea(x,y) ){
				pbd->pushStoneWithState(col,x,y);
				return 1;
			}else{
				return 0;
			}
		}
		
		template<int RINK_ONLY=1,class board_t>
		int simulateSoloF(board_t *const pbd, const fpn_t v, const fpn_t vtheta, const int spin){
			//ただ停止位置を計算し、石を置くだけ
			//元々石がある場合にはおかしなことになりうるので、注意
			const int col=pbd->getTurnColor();
			fpn_t x(FX_THROW),y(FY_THROW),dx,dtheta;
			
			int id=FVtoID(v);
			if( spin==Spin::RIGHT ){
				dx=+V_ALLTable[id].x;
				dtheta=+V_ALLTable[id].theta;
			}else{
				dx=-V_ALLTable[id].x;
				dtheta=-V_ALLTable[id].theta;
			}
			rotateToAdd( V_ALLTable[id].y, dx, 
						vtheta + dtheta,
						&y, &x
						);
			
			if( (!RINK_ONLY) || isInPlayArea(x,y) ){
				pbd->pushStoneWithState(col,x,y);
				return 1;
			}else{
				return 0;
			}
		}
		
		template<int RINK_ONLY=1,class board_t,class mst_t>
		int loopHitZone1_1( board_t *const pbd, const mst_t& arg ){
			//シミュレーション
			//board_tはThinkBoardを想定
			//動いている石1つ、止まっている石1つの場合
			ASSERT( pbd->NActive[0]+pbd->NActive[1]==1, cerr<<pbd->NActive[0]<<","<<pbd->NActive[1]<<endl; );
			
			FMoment aw,as;
			const int aw_col=pbd->getTurnColor();
			const int as_col=pbd->NActive[BLACK]?BLACK:WHITE;
			
			aw.set( arg.x, arg.y, XYtoT(arg.vy,arg.vx), 0, min(FV_MAX,XYtoR(arg.vx,arg.vy)), arg.w );
			
			as.x=pbd->stone[as_col][0].getX();
			as.y=pbd->stone[as_col][0].getY();
			
			fpn_t r_goal=FVtoFR(aw.v);
			uint32 id=FVtoID(aw.v);
			
			fpn_t r_asleep,r_asleep_last;
			
			r_asleep=XYtoR( aw.x - as.x, aw.y - as.y ) - 2*F_STONE_RAD;
			
			for(;;){
				
				//まず非衝突判定
				if( r_goal < r_asleep ){//到達しない
					stepToStop( id, &aw );
					if( (!RINK_ONLY) || isInPlayArea(aw.x,aw.y) ){
						pbd->pushStoneWithState( aw_col, aw.x, aw.y );
						return 1;
					}else{
						return 0;
					}
				}else if( r_asleep < FR_COLLISION ){//衝突
					//衝突処理
					//CERR<<"collision!"<<endl;
					
					collisionAw_As( &aw, &as );
					
					//CERR<<" AW "<<aw.x<<","<<aw.y<<","<<aw.v*sin(aw.theta)<<","<<aw.v*cos(aw.theta)<<","<<aw.w<<endl;
					//CERR<<" AS "<<as.x<<","<<as.y<<","<<as.v*sin(as.theta)<<","<<as.v*cos(as.theta)<<","<<as.w<<endl;
					
					//1_1では1度衝突したら2度と衝突することはない
					stepToStop( FVtoID(aw.v), &aw );
					stepToStop( FVtoID(as.v), &as );
					//フリーガード違反判定
					if( pbd->state[as_col][0].isFreeGuard() ){
						if( RINK_ONLY && !isInPlayArea(as.x,as.y) ){//違反
							DERR<<"freeguard foul 1-1"<<endl;
							return 0;
						}
						pbd->initStonePos();
						pbd->pushStone( as_col, as.x, as.y );
					}else{
						pbd->initStonePos();
						if( (!RINK_ONLY) || isInPlayArea(as.x,as.y) ){
							pbd->pushStone( as_col, as.x, as.y );
						}
					}
					if( (!RINK_ONLY) || isInPlayArea(aw.x,aw.y) ){
						pbd->pushStone( aw_col, aw.x, aw.y );
					}
					return 2;
				}else{//1ステップ進める
					//fpn_t nr=r_goal - r_asleep * 0.8;
					fpn_t nr=sqrt( r_goal*r_goal + r_asleep*r_asleep -2.0*r_goal*r_asleep*cos(FTHETA_CURL_MAX) );
					
					uint32 nid=FRtoID( nr );
					
					//CERR<<aw.x<<","<<aw.y<<","<<aw.v*sin(aw.theta)<<","<<aw.v*cos(aw.theta)<<","<<aw.t;
					step( id, nid, &aw );
					//CERR<<" -> "<<aw.x<<","<<aw.y<<","<<aw.v*sin(aw.theta)<<","<<aw.v*cos(aw.theta)<<","<<aw.t<<endl;
					
					id=nid;
					r_goal=nr;
				}
				
				r_asleep_last=r_asleep;
				r_asleep=XYtoR( aw.x - as.x, aw.y - as.y ) - 2*F_STONE_RAD;
				if( r_asleep > r_asleep_last+0.00001 ){
					r_asleep=9999;
				}
			}
			assert(0);
		}
		
		template<int RINK_ONLY=1,class board_t,class fsbd_t>
		int loopHitZone( board_t *const pabd, fsbd_t *const pbd ){
			//シミュレーション
			//board_tはFSBoardを想定
			//動いている石2つ以上、止まっている石2つ以上の場合
			
			int& NAwake=pbd->NAwake;
			int& NAsleep=pbd->NAsleep;
			BitSet32& awake=pbd->awake;
			BitSet32& asleep=pbd->asleep;
			FSMovingStone *const stone=pbd->stone;
			
			int loop;
			for(loop=0;;++loop){
				
			LSTART:
				assert( pbd->exam() );
				
				if( loop > 100 ){
					//ハマったのでここでBox2Dに投げる
					DERR<<"to box2D"<<endl;
					static StaticCounter c;
					c.add();
					simulateF_b2d_FSB<RINK_ONLY>( pbd, pabd, g_timeStep );
					return 2;
				}
				
				bool coll=false;//衝突が起きうるかどうか
				
				fpn_t minCollT=9999;//衝突しうる再短時間
				
				//awakeなそれぞれの石に対して、asleepな石に対しての移動可能距離を求める
				for( BitSet32 tmp=awake; tmp.any(); tmp.pop_lsb()){
					int aw=tmp.bsf();
					
					//asleep_stones
					const fpn_t r_goal=FVtoFR(stone[aw].v);
					const fpn_t r2_hantei=( r_goal + 2*F_STONE_RAD )*( r_goal + 2*F_STONE_RAD );
					fpn_t r2_asleep=9999;
					//int as_near;
					for( BitSet32 tas=stone[aw].asleep; tas.any(); tas.pop_lsb() ){
						int last=tas.bsf();
						fpn_t r2=XYtoR2( stone[last].x-stone[aw].x, stone[last].y-stone[aw].y );
						if( r2 < r2_hantei ){
							if( r2 < pow(FR_COLLISION+2*F_STONE_RAD,2) ){
								DERR<<"collision AW_AS "<<endl;
								collisionAw_As( &stone[aw], &stone[last] );//衝突
								
								asleep.reset(last);
								
								stone[aw].asleep=asleep;
								stone[last].asleep=asleep;
								
								++NAwake;--NAsleep;
								
								//この2つ以外の石のリストを更新
								BitSet32 taw=awake;
								taw.reset(aw);
								
								stone[aw].awake=taw;
								stone[last].awake=taw;
								
								for(; taw.any(); taw.pop_lsb()){
									int an_aw=taw.bsf();
									stone[an_aw].awake.set(aw);
									stone[an_aw].awake.set(last);
									stone[an_aw].asleep.reset(last);
								}
								
								awake.set(last);
								++loop;
								
								assert( pbd->exam() );
								
								goto LSTART;
							}
							if( r2 < r2_asleep ){
								r2_asleep=r2;
								//as_near=last;
							}
						}else{
							//非衝突が証明された
							stone[aw].asleep.reset(last);
						}
					}
					
					stone[aw].t_goal=FVtoFT(stone[aw].v);
					//CERR<<stone[aw].t_goal<<endl;
					stone[aw].r=r_goal;
					
					if( r2_asleep < r2_hantei ){//静止まで動かせない
						
						fpn_t r_asleep=sqrt( r2_asleep ) - 2*F_STONE_RAD;
						
						fpn_t nr=sqrt( r_goal*r_goal + r_asleep*r_asleep -2.0*r_goal*r_asleep*cos(FTHETA_CURL_MAX) );
						fpn_t mt=stone[aw].t_goal - FRtoFT( nr );
						
						minCollT=min(minCollT,mt);
						
						coll=true;
					}
					
					//awake_stones
					if( stone[aw].awake ){
						//自分以下の番号の石のみ抽出
						BitSet32 taw=stone[aw].awake & ((1U<<aw)-1U);
						
						//CERR<<taw.count()<<endl;
						
						for( ; taw.any(); taw.pop_lsb() ){
							int last=taw.bsf();
							
							fpn_t r2=XYtoR2( stone[aw].x - stone[last].x, stone[aw].y - stone[last].y );
							
							if( r2 < pow(FR_COLLISION+2*F_STONE_RAD,2) ){
								//動-動衝突
								DERR<<"collision 2AW"<<endl;
								collision2Aw( &stone[aw], &stone[last] );//衝突
								//動-動関係を整理
								if( NAwake > 2 ){
									BitSet32 ttaw=awake;
									ttaw.reset(aw);ttaw.reset(last);
						
									stone[aw].awake=ttaw;
									stone[last].awake=ttaw;
									
									for( ; ttaw.any(); ttaw.pop_lsb() ){
										int an_aw=ttaw.bsf();
										stone[an_aw].awake.set(aw);
										stone[an_aw].awake.set(last);
									}
								}else{
									stone[aw].awake.reset();
									stone[last].awake.reset();
								}
								++loop;
								
								goto LSTART;
							}
							
							fpn_t r_awake=sqrt( r2 ) -2*F_STONE_RAD;
							
							if( r_awake < r_goal + stone[last].r ){//静止まで動かせない
								
								//減速せずに互いに一直線に近づくときの衝突までの時間を計算
								fpn_t t=r_awake / ( stone[aw].v + stone[last].v );
								
								minCollT=min(minCollT,t);
								coll=true;
							}else{//非衝突が証明された
								stone[aw].awake.reset(last);
								stone[last].awake.reset(aw);
							}
						}
					}
				}
				
				//全ての石が衝突せずに静止する事が証明されたので、全ての石を置いて終わり
				if( !coll ){
					DERR<<"no collision"<<endl;
					for( BitSet32 taw=awake; taw.any(); taw.pop_lsb() ){
						int aw=taw.bsf();
						stepToStop( FVtoID(stone[aw].v), &stone[aw] );
						if( (!RINK_ONLY) || isInPlayArea(stone[aw].x,stone[aw].y) ){//エリア内
							asleep.set(aw);
						}else{//エリア外
							if( stone[aw].state.isFreeGuard() ){//フリーガード違反
								DERR<<"freeguard foul"<<endl;
								return 0;
							}
						}
					}
					goto END;
				}
				
				DERR<<"Min Time : "<<minCollT<<endl;
				
				//それぞれの石を移動可能な分動かす
				{
					for( BitSet32 taw=awake; taw.any(); taw.pop_lsb() ){
						int aw=taw.bsf();
						
						DERR<<"("<<aw<<") "<<stone[aw].x<<","<<stone[aw].y<<","
						<<stone[aw].v*sin(stone[aw].theta)<<","
						<<stone[aw].v*cos(stone[aw].theta)<<","<<stone[aw].t;
						
						if( stone[aw].t_goal < minCollT ){
							//最短衝突時刻より停止時刻が早い石は静止させる
							
							stepToStop( FVtoID(stone[aw].v), &stone[aw] );
							
							DERR<<" -> "<<stone[aw].x<<","<<stone[aw].y<<","
							<<stone[aw].v*sin(stone[aw].theta)<<","
							<<stone[aw].v*cos(stone[aw].theta)<<","<<stone[aw].t<<" stop"<<endl;
							
							awake.reset(aw);
							--NAwake;
							
							if( (!RINK_ONLY) || isInPlayArea(stone[aw].x,stone[aw].y) ){//エリア内
								asleep.set(aw);
								++NAsleep;
								for( BitSet32 ttaw=awake; ttaw.any(); ttaw.pop_lsb() ){
									int an_aw=ttaw.bsf();
									stone[an_aw].awake.reset(aw);
									stone[an_aw].asleep.set(aw);
								}
							}else{//エリア外
								if( stone[aw].state.isFreeGuard() ){//フリーガード違反
									DERR<<"freeguard foul"<<endl;
									return 0;
								}
								for( BitSet32 ttaw=awake; ttaw.any(); ttaw.pop_lsb() ){
									int an_aw=ttaw.bsf();
									stone[an_aw].awake.reset(aw);
								}
							}
						}else{
							int nid=FVtoID(stone[aw].v - FTtoFV(minCollT));
							if( nid>0 ){
								//静止しない
								step( FVtoID(stone[aw].v), nid, &stone[aw] );
								
								DERR<<" -> "<<stone[aw].x<<","<<stone[aw].y<<","
								<<stone[aw].v*sin(stone[aw].theta)<<","
								<<stone[aw].v*cos(stone[aw].theta)<<","<<stone[aw].t<<endl;
								
								if( RINK_ONLY && !isInPlayArea(stone[aw].x,stone[aw].y) ){//エリア外
									if( stone[aw].state.isFreeGuard() ){//フリーガード違反
										DERR<<"freeguard foul"<<endl;
										return 0;
									}
									awake.reset(aw);
									--NAwake;
									for( BitSet32 ttaw=awake; ttaw.any(); ttaw.pop_lsb() ){
										int an_aw=ttaw.bsf();
										stone[an_aw].awake.reset(aw);
									}
								}
							}else{//丸め込みにより静止
								stepToStop( FVtoID(stone[aw].v), &stone[aw] );
								
								DERR<<" -> "<<stone[aw].x<<","<<stone[aw].y<<","
								<<stone[aw].v*sin(stone[aw].theta)<<","
								<<stone[aw].v*cos(stone[aw].theta)<<","<<stone[aw].t<<" stop"<<endl;
								
								awake.reset(aw);
								--NAwake;
								
								if( (!RINK_ONLY) || isInPlayArea(stone[aw].x,stone[aw].y) ){//エリア内
									asleep.set(aw);
									++NAsleep;
									for( BitSet32 ttaw=awake; ttaw.any(); ttaw.pop_lsb() ){
										int an_aw=ttaw.bsf();
										stone[an_aw].awake.reset(aw);
										stone[an_aw].asleep.set(aw);
									}
								}else{//エリア外
									if( stone[aw].state.isFreeGuard() ){//フリーガード違反
										DERR<<"freeguard foul"<<endl;
										return 0;
									}
									for( BitSet32 ttaw=awake; ttaw.any(); ttaw.pop_lsb() ){
										int an_aw=ttaw.bsf();
										stone[an_aw].awake.reset(aw);
									}
								}
							}
						}
					}
				}
				
				if( NAwake==0 ){
					break;
				}
			}
		END:
			//プレイアウト用の局面構造体に石を置く
			pabd->initStonePos();
			for( BitSet32 tmp=asleep; tmp.any() ;tmp.pop_lsb()){
				int id=tmp.bsf();
				DERR<<id<<" ";
				pabd->pushStone( stone[id].state.getColor(), stone[id].x, stone[id].y );
			}
			DERR<<endl;
			
			DERR<<loop<<endl;
			
			return 2;
		}
		
		template<int RINK_ONLY=1,class board_t,class mst_t>
		int loopHitZone1_N( board_t *const pbd, const mst_t& arg ){
			//ヒットゾーンでのシミュレーション
			//board_tはThinkBoardを想定
			//動いている石1つ、止まっている石2つ以上の場合
			struct AsInfo{
				int i;
				fpn_t r2_last;
			};
			
			AsInfo asleepList[16];
			const int aw_col=pbd->getTurnColor();
			
			fpn_t v_goal=min(FV_MAX,XYtoR(arg.vx,arg.vy));
			fpn_t r_goal=FVtoFR(v_goal);
			fpn_t r2_hantei=( r_goal + 2*F_STONE_RAD )*( r_goal + 2*F_STONE_RAD );
			fpn_t r2_asleep=9999;
			int as_near;
			
			//asleepList作成
			fpn_t vtheta=XYtoT( arg.vy, arg.vx );
			fpn_t dvtheta;
			if( arg.w > 0 ){
				dvtheta=+FTHETA_CURL_MAX/2;
			}else{
				dvtheta=-FTHETA_CURL_MAX/2;
			}
			fpn_t vtheta_ex=vtheta+dvtheta;
			int NAsleep=0;
			for(int c=0;c<2;++c){
				for(int n=0;n<pbd->NActive[c];++n){
					//衝突可能性判定
					fpn_t dx=pbd->stone[c][n].getX() - arg.x;
					fpn_t dy=pbd->stone[c][n].getY() - arg.y;
					fpn_t r2=XYtoR2( dx, dy );
					if( r2 < r2_hantei ){
						fpn_t theta=XYtoT( dy, dx );
						//if( fabs( theta-vtheta_ex ) <= FTHETA_CURL_MAX/2 ){
							int as_num=c*8 + n;
							asleepList[NAsleep].i=as_num;
							asleepList[NAsleep].r2_last=r2;
							++NAsleep;
							if( r2<r2_asleep ){
								r2_asleep=r2;
								as_near=as_num;
							}
						//}
					}
				}
			}
			
			//cerr<<NAsleep<<endl;
			
			FMoment aw;
			aw.set( arg.x, arg.y, vtheta, 0, v_goal, arg.w );
			uint32 id=FVtoID(v_goal);
			
			for(;;){
				if( r2_asleep > r2_hantei ){//到達しない
					stepToStop( id, &aw );
					if( (!RINK_ONLY) || isInPlayArea(aw.x,aw.y) ){
						pbd->pushStoneWithState( aw_col, aw.x, aw.y );
						return 1;
					}else{
						return 0;
					}
				}else if( r2_asleep < pow(FR_COLLISION + 2*F_STONE_RAD,2)  ){//衝突
					FMoment as;
					int c=as_near/8;
					int n=as_near%8;
					as.x = pbd->stone[c][n].x;
					as.y = pbd->stone[c][n].y;
					collisionAw_As( &aw, &as );
					FSBoard fsb;
					fsb.setAfterFirstColl( *pbd, as_near, aw, as );
					return loopHitZone<RINK_ONLY>( pbd, &fsb );
					
					//return simulateF_b2d<RINK_ONLY>( pbd, aw, g_timeStep );
				}else{//1ステップ進める
					fpn_t r_asleep=sqrt(r2_asleep)-2*F_STONE_RAD;
					
					//fpn_t nr=r_goal - r_asleep * 0.8;
					fpn_t nr=sqrt( r_goal*r_goal + r_asleep*r_asleep -2.0*r_goal*r_asleep*cos(FTHETA_CURL_MAX) );
					uint32 nid=FRtoID( nr );
					//CERR<<aw.x<<","<<aw.y<<","<<aw.v*sin(aw.theta)<<","<<aw.v*cos(aw.theta)<<","<<aw.t;
					step( id, nid, &aw );
					//CERR<<" -> "<<aw.x<<","<<aw.y<<","<<aw.v*sin(aw.theta)<<","<<aw.v*cos(aw.theta)<<","<<aw.t<<endl;
					id=nid;
					r_goal=nr;
				}
				//最も近距離の石を探す
				r2_hantei=( r_goal + 2*F_STONE_RAD )*( r_goal + 2*F_STONE_RAD );
				r2_asleep=9999;
				vtheta_ex=aw.theta+dvtheta;
				for(int i=NAsleep-1;i>=0;--i){
					int c=asleepList[i].i/8;
					int n=asleepList[i].i%8;
					fpn_t dx=pbd->stone[c][n].getX() - aw.x;
					fpn_t dy=pbd->stone[c][n].getY() - aw.y;
					fpn_t r2=XYtoR2( dx, dy );
					if( r2<r2_hantei && r2 < asleepList[i].r2_last + 0.00001 ){
						
						fpn_t theta=XYtoT( dy, dx );
						//if( fabs( theta-vtheta_ex ) <= FTHETA_CURL_MAX/2 ){
						//衝突可能性あり
						//ASSERT( !((sqrt(r2)-2*F_STONE_RAD < FR_COLLISION) && (r2>=asleepList[i].r2_last)),
						//	   cerr<<r2<<" , "<<asleepList[i].r2_last<<endl;);
						
						asleepList[i].r2_last=r2;
						if( r2<r2_asleep ){
							r2_asleep=r2;
							as_near=asleepList[i].i;
						}
						continue;
						//}
					}
					//非衝突が証明された
					asleepList[i]=asleepList[--NAsleep];//リストを詰める
				}
			}
			assert(0);
		}
		
		template<class board_t,class move_t>
		int simulateF( board_t *const pbd, const move_t& mv )
		{
			//高速シミュレータ
			int ret;
			
			//シミュレーション開始座標と速度を求める
			
			//停止距離計算
			fpn_t v=min( FV_MAX, XYtoR(mv.getVX(),mv.getVY()) );
			if( FVtoFR(v) < pbd->frontY - F_STONE_RAD - FY_THROW ){
				//どの石にも届かず停止する
				return simulateSoloF( pbd, v, XYtoT(mv.getVY(),mv.getVX()), mv.getSpin() );
			}else{
				
				fMStone<fpn_t> org;
				org.set( FX_THROW, FY_THROW, mv.getVX(), mv.getVY(), mv.isLeftSpin()?(-FANGV_ORG):FANGV_ORG );
				
				if( pbd->NActive[BLACK]+pbd->NActive[WHITE]==1 ){
					//tick();
					ret=loopHitZone1_1<1>( pbd, org );
					//tock();
				}else{
					//tick();
					ret=loopHitZone1_N<1>( pbd, org );
					//tock();
				}
			}
			return ret;
		}
		
		template<int SPIN,class pos_t>
		void FV_FRtoFXY(fpn_t v,fpn_t r,pos_t *const pos,fpn_t *const theta){
			//初速vのショットが投擲点から距離rの点での相対位置を求める
			//もしrまで到達しないならば、適当な値が帰るので注意
			fpn_t fv=min(v,FV_MAX);
			int id=FVtoID(fv);
			fpn_t gx,gy,x,y;
			if( SPIN==Spin::RIGHT ){
				*theta=-V_ALLTable[id].theta;
				gx=V_ALLTable[id].x;
				gy=V_ALLTable[id].y;
			}else{
				*theta=V_ALLTable[id].theta;
				gx=-V_ALLTable[id].x;
				gy=V_ALLTable[id].y;
			}
			
			//到達点から距離fr-rの点を探す
			fpn_t fr=FVtoFR( fv );
			if( fr-r > FR_COLLISION ){
				
				fpn_t r2_coll=r*(r+2*FR_COLLISION);
				int cnt=0;
				
				fpn_t tr=fr-r;
				int nid=FRtoID(tr);
				
				while(1){
					//CERR<<*pos<<endl;
					//fpn_t nv=FRtoFV()
					//CERR<<fr<<" -> ";
					//この点のところまで位置を移動
					if( SPIN==Spin::RIGHT ){
						x=gx-V_ALLTable[nid].x;
						y=gy-V_ALLTable[nid].y;
						fr=XYtoR2( x, y );
					}else{
						x=gx+V_ALLTable[nid].x;
						y=gy-V_ALLTable[nid].y;
						fr=XYtoR2( x, y );
					}
					//CERR<<sqrt(fr)<<" ( "<<r<<" )"<<" nid "<<nid<<endl;
					++cnt;
					
					if( fr < r2_coll ){
						//ASSERT( fr>r*r, cerr<<sqrt(fr)<<" ( "<<r<<" )"<<endl; );
						break;
					}
					fr=sqrt(fr);
					int nnid;
					do{
						tr+=fr-r;
						//CERR<<tr<<endl;
						nnid=FRtoID(tr);
					}while( nnid==nid );
					nid=nnid;
				}
				//CERR<<cnt<<endl;
			}
			
			pos->setX( x );pos->setY( y );
			//CERR<<" ( "<<x<<" , "<<y<<" )"<<endl;
		}
		
		template<class pos_t,class move_t>
		void rotateToPassPointF(const pos_t& pos,fpn_t v,move_t *const mv){
			//指定された初速で、指定された位置を通る速度ベクトルを生成
			
			//tick();
			
			const fPosXY<> tar( pos.x - FX_THROW, pos.y - FY_THROW );
			const fpn_t r=XYtoR( tar.getX(), tar.getY() );
			
			fPosXY<> tp;
			fpn_t theta;
			
			if( mv->isLeftSpin() ){
				FV_FRtoFXY<Spin::LEFT>( v, r, &tp ,&theta );
			}else{
				FV_FRtoFXY<Spin::RIGHT>( v, r, &tp ,&theta );
			}
			fpn_t dtheta=XYtoT( tar.y, tar.x ) - XYtoT( tp.y, tp.x ) + theta;
			rotate( v, 0, dtheta, &mv->y, &mv->x );
			
			//tock();
			//CERR<<mv->x<<" , "<<mv->y<<endl;
			
		}
		
		template<class pos_t,class move_t>
		void rotateToPassPointGoF(const pos_t& pos,fpn_t r,move_t *const mv){
			//指定された位置を通り、さらに指定距離分(投擲点から見て)進む速度ベクトルを生成
			const fPosXY<> tar( pos.x - FX_THROW, pos.y - FY_THROW );
			const fpn_t pr=XYtoR( tar.getX(), tar.getY() );
			const fpn_t v=FRtoFV(pr+r);
			
			fPosXY<> tp;
			fpn_t theta;
			
			if( mv->isLeftSpin() ){
				FV_FRtoFXY<Spin::LEFT>( v, pr, &tp ,&theta );
			}else{
				FV_FRtoFXY<Spin::RIGHT>( v, pr, &tp ,&theta );
			}
			fpn_t dtheta=XYtoT( tar.y, tar.x ) - XYtoT( tp.y, tp.x ) + theta;
			rotate( v, 0, dtheta, &mv->y, &mv->x );
			
			//CERR<<mv->x<<" , "<<mv->y<<endl;
		}
		
	}
}